@@ -8,11 +8,17 @@ these are not really related, formatting could be factored out as
 a separate base
 """
 
+import sys
 import copy
 
-from collections import MutableSet, Sized, Set
 
-from ruamel.yaml.compat import ordereddict, PY2
+from ruamel.yaml.compat import ordereddict, PY2, string_types
+from ruamel.yaml.scalarstring import ScalarString
+
+if PY2:
+    from collections import MutableSet, Sized, Set
+else:
+    from collections.abc import MutableSet, Sized, Set
 
 if False:  # MYPY
     from typing import Any, Dict, Optional, List, Union  # NOQA
@@ -175,6 +181,10 @@ class Tag(object):
         # type: () -> None
         self.value = None
 
+    def __repr__(self):
+        # type: () -> Any
+        return '{0.__class__.__name__}({0.value!r})'.format(self)
+
 
 class CommentedBase(object):
     @property
@@ -194,25 +204,25 @@ class CommentedBase(object):
 
     def yaml_key_comment_extend(self, key, comment, clear=False):
         # type: (Any, Any, bool) -> None
-        l = self.ca._items.setdefault(key, [None, None, None, None])
-        if clear or l[1] is None:
+        r = self.ca._items.setdefault(key, [None, None, None, None])
+        if clear or r[1] is None:
             if comment[1] is not None:
                 assert isinstance(comment[1], list)
-            l[1] = comment[1]
+            r[1] = comment[1]
         else:
-            l[1].extend(comment[0])
-        l[0] = comment[0]
+            r[1].extend(comment[0])
+        r[0] = comment[0]
 
     def yaml_value_comment_extend(self, key, comment, clear=False):
         # type: (Any, Any, bool) -> None
-        l = self.ca._items.setdefault(key, [None, None, None, None])
-        if clear or l[3] is None:
+        r = self.ca._items.setdefault(key, [None, None, None, None])
+        if clear or r[3] is None:
             if comment[1] is not None:
                 assert isinstance(comment[1], list)
-            l[3] = comment[1]
+            r[3] = comment[1]
         else:
-            l[3].extend(comment[0])
-        l[2] = comment[0]
+            r[3].extend(comment[0])
+        r[2] = comment[0]
 
     def yaml_set_start_comment(self, comment, indent=0):
         # type: (Any, Any) -> None
@@ -238,19 +248,21 @@ class CommentedBase(object):
         from ruamel.yaml.tokens import CommentToken
 
         def comment_token(s, mark):
-            # type: (Any) -> Any
+            # type: (Any, Any) -> Any
             # handle empty lines as having no comment
             return CommentToken(('# ' if s else '') + s + '\n', mark, None)
 
         if after_indent is None:
             after_indent = indent + 2
-        if before and before[-1] == '\n':
+        if before and (len(before) > 1) and before[-1] == '\n':
             before = before[:-1]  # strip final newline if there
         if after and after[-1] == '\n':
             after = after[:-1]  # strip final newline if there
         start_mark = CommentMark(indent)
         c = self.ca.items.setdefault(key, [None, [], None, None])
-        if before:
+        if before == '\n':
+            c[1].append(comment_token('', start_mark))
+        elif before:
             for com in before.split('\n'):
                 c[1].append(comment_token(com, start_mark))
         if after:
@@ -258,7 +270,7 @@ class CommentedBase(object):
             if c[3] is None:
                 c[3] = []
             for com in after.split('\n'):
-                c[3].append(comment_token(com, start_mark))
+                c[3].append(comment_token(com, start_mark))  # type: ignore
 
     @property
     def fa(self):
@@ -358,6 +370,10 @@ class CommentedBase(object):
         # type: () -> Any
         raise NotImplementedError
 
+    def _yaml_get_column(self, key):
+        # type: (Any) -> Any
+        raise NotImplementedError
+
 
 class CommentedSeq(list, CommentedBase):
     __slots__ = Comment.attrib,
@@ -384,23 +400,23 @@ class CommentedSeq(list, CommentedBase):
         for list_index in sorted(self.ca.items, reverse=True):
             if list_index < idx:
                 break
-            self.ca.items[list_index+1] = self.ca.items.pop(list_index)
+            self.ca.items[list_index + 1] = self.ca.items.pop(list_index)
 
     def pop(self, idx=None):
         # type: (Any) -> Any
-        res = list.pop(self, idx)
+        res = list.pop(self, idx)  # type: ignore
         self.ca.items.pop(idx, None)  # might not be there -> default value
         for list_index in sorted(self.ca.items):
             if list_index < idx:
                 continue
-            self.ca.items[list_index-1] = self.ca.items.pop(list_index)
+            self.ca.items[list_index - 1] = self.ca.items.pop(list_index)
         return res
 
     def _yaml_get_column(self, key):
         # type: (Any) -> Any
         column = None
         sel_idx = None
-        pre, post = key-1, key+1
+        pre, post = key - 1, key + 1
         if pre in self.ca.items:
             sel_idx = pre
         elif post in self.ca.items:
@@ -428,13 +444,23 @@ class CommentedSeq(list, CommentedBase):
 
     def __deepcopy__(self, memo):
         # type: (Any) -> Any
-        res = CommentedSeq()
+        res = self.__class__()
         memo[id(self)] = res
         for k in self:
             res.append(copy.deepcopy(k))
             self.copy_attributes(res, deep=True)
         return res
 
+    def __setitem__(self, idx, value):
+        # type: (Any, Any) -> None
+        # try to preserve the scalarstring type if setting an existing key to a new value
+        if idx < len(self):
+            if isinstance(value, string_types) and \
+               not isinstance(value, ScalarString) and \
+               isinstance(self[idx], ScalarString):
+                value = type(self[idx])(value)
+        list.__setitem__(self, idx, value)
+
 
 class CommentedKeySeq(tuple, CommentedBase):
     """This primarily exists to be able to roundtrip keys that are sequences"""
@@ -457,7 +483,7 @@ class CommentedKeySeq(tuple, CommentedBase):
         # type: (Any) -> Any
         column = None
         sel_idx = None
-        pre, post = key-1, key+1
+        pre, post = key - 1, key + 1
         if pre in self.ca.items:
             sel_idx = pre
         elif post in self.ca.items:
@@ -662,8 +688,8 @@ class CommentedMap(ordereddict, CommentedBase):
             if level >= len(key_list):
                 if level > len(key_list):
                     raise IndexError
-                return d[key_list[level-1]]
-            return get_one_level(key_list, level+1, d[key_list[level-1]])
+                return d[key_list[level - 1]]
+            return get_one_level(key_list, level + 1, d[key_list[level - 1]])
 
         try:
             return get_one_level(key, 1, self)
@@ -684,6 +710,16 @@ class CommentedMap(ordereddict, CommentedBase):
                     return merged[1][key]
             raise
 
+    def __setitem__(self, key, value):
+        # type: (Any, Any) -> None
+        # try to preserve the scalarstring type if setting an existing key to a new value
+        if key in self:
+            if isinstance(value, string_types) and \
+               not isinstance(value, ScalarString) and \
+               isinstance(self[key], ScalarString):
+                value = type(self[key])(value)
+        ordereddict.__setitem__(self, key, value)
+
     def _unmerged_contains(self, key):
         # type: (Any) -> Any
         if ordereddict.__contains__(self, key):
@@ -704,7 +740,7 @@ class CommentedMap(ordereddict, CommentedBase):
         # type: (Any, Any) -> Any
         try:
             return self.__getitem__(key)
-        except:
+        except:  # NOQA
             return default
 
     def __repr__(self):
@@ -821,14 +857,14 @@ class CommentedMap(ordereddict, CommentedBase):
             yield x, ordereddict.__getitem__(self, x)
         done = []  # type: List[Any]  # list of processed merge items, kept for masking
         for merged in getattr(self, merge_attrib, []):
-            for x in merged[1]:
+            for x, v in merged[1].items():
                 if ordereddict.__contains__(self, x):
                     continue
                 for y in done:
                     if x in y:
                         break
                 else:
-                    yield x, ordereddict.__getitem__(merged[1], x)
+                    yield x, v  # ordereddict.__getitem__(merged[1], x)
             done.append(merged[1])
 
     if PY2:
@@ -861,7 +897,7 @@ class CommentedMap(ordereddict, CommentedBase):
 
     def __deepcopy__(self, memo):
         # type: (Any) -> Any
-        res = CommentedMap()
+        res = self.__class__()
         memo[id(self)] = res
         for k in self:
             res[k] = copy.deepcopy(self[k])
@@ -909,3 +945,30 @@ class CommentedSet(MutableSet, CommentedMap):
     def __repr__(self):
         # type: () -> str
         return 'set({0!r})'.format(self.odict.keys())
+
+
+class TaggedScalar(CommentedBase):
+    # the value and style attributes are set during roundtrip construction
+    def __str__(self):
+        return self.value
+
+
+def dump_comments(d, name='', sep='.', out=sys.stdout):
+    # type: (Any, str, str, Any) -> None
+    """
+    recursively dump comments, all but the toplevel preceded by the path
+    in dotted form x.0.a
+    """
+    if isinstance(d, dict) and hasattr(d, 'ca'):
+        if name:
+            print(name)
+        print(d.ca, file=out)  # type: ignore
+        for k in d:
+            dump_comments(d[k], name=(name + sep + k) if name else k, sep=sep, out=out)
+    elif isinstance(d, list) and hasattr(d, 'ca'):
+        if name:
+            print(name)
+        print(d.ca, file=out)  # type: ignore
+        for idx, k in enumerate(d):
+            dump_comments(k, name=(name + sep + str(idx)) if name else str(idx),
+                          sep=sep, out=out)
